/**
 * Any CSS included here will be global. The classic template
 * bundles Infima by default. Infima is a CSS framework designed to
 * work well for content-centric websites.
 */

/* You can override the default Infima variables here. */
:root {
  --ifm-color-primary: #E8710A;
  --ifm-color-primary-dark: #d16609;
  --ifm-color-primary-darker: #ba5a08;
  --ifm-color-primary-darkest: #ba5a08;
  --ifm-color-primary-light: #f57e18;
  --ifm-color-primary-lighter: #f68d31;
  --ifm-color-primary-lightest: #f79b4b;
  --ifm-code-font-size: 95%;
  --docusaurus-highlighted-code-line-bg: rgba(0, 0, 0, 0.1);
}

/* For readability concerns, you should choose a lighter palette in dark mode. */
[data-theme='dark'] {
  --ifm-color-primary: #F9AB00;
  --ifm-color-primary-dark: #e09a00;
  --ifm-color-primary-darker: #c78900;
  --ifm-color-primary-darkest: #ae7800;
  --ifm-color-primary-light: #ffb514;
  --ifm-color-primary-lighter: #ffbe2e;
  --ifm-color-primary-lightest: #F9AB00;
  --docusaurus-highlighted-code-line-bg: rgba(0, 0, 0, 0.3);
}
